function [holdL, holdk, holdnumbclust, holdM, holdpred, pred] = sampler_DP(data, M, updateM, sampler, burnin, every, numbofits, predvalues)

% Calculating predictive distribution or not?
if ( nargin == 7 )
    preds = 0;
else
    preds = 1;
    pred = zeros(1, length(predvalues));
end

% Setting hyperparameters for the mixture model components
mu0 = mean(data);
sigmasq0 = 30^2;
sigmasqalpha = 1;
sigmasqbeta = 1;

% Set up total number of iterations
totalits = burnin + every * numbofits;

L = 10^(-2);

% Set up the holdtheta matrix to store samples of M - empty if updateM parameters is zero
if ( updateM == 1 )
    holdM = zeros(1, numbofits);
else
    holdM = [];
end

% Set up matrices to store results
holdL = zeros(1, numbofits);
holdk = zeros(1, numbofits);
holdnumbclust = zeros(1, numbofits);
holdpred = zeros(1, numbofits);

n = length(data);

% Set up for calculating predictive or not
if (preds == 1)
    s = 1:n;
    sn = n + 1;
    kstar = n + 1;
else
    s = 1:n;
    kstar = n;
end

% Set jump sizes and parameters of components
J = zeros(1, kstar);
mu = mu0+sqrt(sigmasq0)*randn(1, kstar);
sigmasq = 1./gamrnd(sigmasqalpha, 1/sigmasqbeta, 1, kstar);

for i = 1:kstar
    J(i) = draw_DP(L);
end

% Set up slice variables
if ( preds == 1 )
    sumphi = gamrnd(n + 1, 1 / sum(J));
    if ( strcmp(sampler, 'slice1') == 1 )
        ustar = L + rand(1, n + 1) .* ([J(s) J(sn)] - L);
        L = min(ustar);
    else
        ustar = rand * min([J(s) J(sn)]);
        L = ustar;
    end
else
    sumphi = gamrnd(n, 1 / sum(J));
    if ( strcmp(sampler, 'slice1') == 1 )
        ustar = L + rand(1, n) .* (J(s) - L);
        L = min(ustar);
    else
        ustar = rand * min(J(s));
        L = ustar;
    end
end

phiaccept = 0;
phicount = 0;

for it = 1:totalits
    
    % Update the latent variable v (with the name sumphi in this code)
    newsumphi = sumphi * exp(0.5 * randn);

    if ( preds == 1)
        newloglike = (n + 1) * log(newsumphi) - newsumphi * sum(J) - M * log(1 + newsumphi) - M * expint(L * (1 + newsumphi));
        loglike = (n + 1) * log(sumphi) - sumphi * sum(J) - M * log(1 + sumphi) - M * expint(L * (1 + sumphi));
    else
        newloglike = n * log(newsumphi) - newsumphi * sum(J) - M * log(1 + newsumphi) - M * expint(L * (1 + newsumphi));
        loglike = n * log(sumphi) - sumphi * sum(J) - M * log(1 + sumphi) - M * expint(L * (1 + sumphi));
    end
    logaccept = newloglike - loglike;

    accept = 1;
    if ( logaccept < 0 )
        accept = exp(logaccept);
    end

    phiaccept = phiaccept + accept;
    phicount = phicount + 1;

    if ( rand < accept )
        sumphi = newsumphi;
    end

    % Update jumps with observations allocated to them
    nj = zeros(1, length(J));
    for j = 1:n
        nj(s(j)) = nj(s(j)) + 1;
    end
    if ( preds == 1 )
        nj(sn) = nj(sn) + 1;
    end

    holds = [];
    counts = zeros(1, length(J));
    counter = 0;
    for j = 1:length(J)
        if ( nj(j) > 0 )
            holds = [holds j];
            counter = counter + 1;
            counts(j) = counter;
            J(j) = gamrnd(nj(j), 1 / (1 + sumphi));
        end
    end

    J = J(holds);
    mu = mu(holds);
    sigmasq = sigmasq(holds);
    s = counts(s);

    % Update parameters of components with observations allocated to them
    for j = 1:length(mu)
        meanstar = (sum(data(s==j)) / sigmasq(j) + mu0 / sigmasq0) / (sum(s==j) / sigmasq(j) + 1 / sigmasq0);
        varstar = 1 / (sum(s==j) / sigmasq(j) + 1 / sigmasq0);

        mu(j) = meanstar + sqrt(varstar) * randn;

        alphastar = sigmasqalpha + 0.5 * sum(s==j);
        betastar = sigmasqbeta + 0.5 * sum((data(s==j) - mu(j)).^2);

        sigmasq(j) = 1 / gamrnd(alphastar, 1 / betastar);
    end

    % Simulate slice variables
    if ( preds == 1 )
        sn = counts(sn);
        if ( strcmp(sampler, 'slice1') == 1 )
            ustar = rand(1, n + 1) .* [J(s) J(sn)];
            L = min(ustar);
        else
            ustar = rand * min([J(s) J(sn)]);
            L = ustar;
        end
    else
        if ( strcmp(sampler, 'slice1') == 1 )
            ustar = rand(1, n) .* J(s);
            L = min(ustar);
        else
            ustar = rand * min(J(s));
            L = ustar;
        end
    end

    % Simulation the jump process without observations allocated to them
    Lstar = (1 + sumphi) * L;
    limit = expint(Lstar);

    kstar = min(10^7, poissrnd(M * limit));
    
    oldlengthJ = length(J);
    J = [J zeros(1, kstar)];
    mu = [mu mu0+sqrt(sigmasq0)*randn(1, kstar)];
    sigmasq = [sigmasq 1./gamrnd(sigmasqalpha, 1 / sigmasqbeta, 1, kstar)];
    for j = (oldlengthJ+1):(oldlengthJ+kstar)
        F =  draw_DP(Lstar);
        J(j) = F / (1 + sumphi);
    end

    % Update allocation variables, s
    if ( strcmp(sampler, 'slice1') == 1 )
        for j = 1:n
            prob = (J>ustar(j)).*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));

            fprob = cumsum(prob);
            fprob = fprob / fprob(end);

            u = rand;
            s(j) = 1;
            while ( fprob(s(j)) < u )
                s(j) = s(j) + 1;
            end
        end
        if (preds == 1)
            prob = J>ustar(sn);

            fprob = cumsum(prob);
            fprob = fprob / fprob(end);

            u = rand;
            sn = 1;
            while ( fprob(sn) < u )
                sn = sn + 1;
            end
        end
    else
        for j = 1:n
            if ( preds == 0 )
                sminus = s([1:(j-1) (j+1):n]);
            else
                sminus = [sn s([1:(j-1) (j+1):n])];
            end
            [Jstar, findJstar] = min(J(sminus));
            findJstar = sminus(findJstar);
            prob = (J>Jstar).*J.*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));
            prob(findJstar) = J(findJstar)*exp(- 0.5 * (data(j) - mu(findJstar))^2 ./ sigmasq(findJstar) - 0.5 * log(sigmasq(findJstar)));
            prob = prob + (J<Jstar).*(J>ustar).*Jstar.*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));

            fprob = cumsum(prob);
            fprob = fprob / fprob(end);

            u = rand;
            s(j) = 1;
            while ( fprob(s(j)) < u )
                s(j) = s(j) + 1;
            end
        end

        if ( preds == 1)
            sminus = s;
            [Jstar, findJstar] = min(J(sminus));
            findJstar = sminus(findJstar);
            prob = (J>Jstar).*J.*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));
            prob(findJstar) = J(findJstar)*exp(- 0.5 * (data(j) - mu(findJstar))^2 ./ sigmasq(findJstar) - 0.5 * log(sigmasq(findJstar)));
            prob = prob + (J<Jstar).*(J>ustar).*Jstar.*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));

            fprob = cumsum(prob);
            fprob = fprob / fprob(end);

            u = rand;
            sn = 1;
            while ( fprob(sn) < u )
                sn = sn + 1;
            end
        end
    end

    % Update M
    if ( updateM == 1 )
        alphastar = 1 + length(J);
        betastar = 1 + log(1 + sumphi) + expint((1 + sumphi) * L);

        M = gamrnd(alphastar, 1 / betastar);
    end

    % Record output of chain
    if ( (it > burnin) && (mod(it - burnin, every) == 0) )
        if (updateM == 1)
            holdM((it - burnin) / every) = M;
        end
        holdL((it - burnin) / every) = L;
        holdk((it - burnin) / every) = length(J);
        holdnumbclust((it - burnin) / every) = length(unique(s));

        % Rao-Blackwellised estimate of predictive density
        if ( preds == 1 )
            holdpred((it - burnin) / every) = mu(sn) + sqrt(sigmasq(sn)) * randn;

            if ( strcmp(sampler, 'slice1') == 1 )
                for j = 1:length(predvalues)
                    prob = (J>ustar(sn)).*exp(- 0.5 * (predvalues(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));

                    pred(j) = pred(j) + sum(prob)/sum(J>ustar(sn));
                end
            else
                sminus = s;
                [Jstar, findJstar] = min(J(sminus));
                findJstar = sminus(findJstar);
                for j = 1:length(predvalues)
                    prob = (J>Jstar).*J.*exp(- 0.5 * (predvalues(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));
                    prob(findJstar) = J(findJstar)*exp(- 0.5 * (predvalues(j) - mu(findJstar))^2 ./ sigmasq(findJstar) - 0.5 * log(sigmasq(findJstar)));
                    prob = prob + (J<Jstar).*(J>ustar).*Jstar.*exp(- 0.5 * (predvalues(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));
                    pred(j) = pred(j) + sum(prob) ./ sum(J(J>ustar));
                end
            end
        end
    end
    
    if ( mod(it, 100) == 0 )
        it
        M
        phiaccept / phicount
        length(unique(s))
    end
end

if (preds == 1)
    pred = pred / numbofits;
else
    pred = [];
    holdpred = [];
end